#******************************************************************************
#
#       Copyright:      2005-2006 Paul Obermeier (obermeier@poSoft.de)
#
#                       See the file "Tcl3D_License.txt" for information on
#                       usage and redistribution of this file, and for a
#                       DISCLAIMER OF ALL WARRANTIES.
#
#       Module:         Tcl3D -> tcl3dUtil
#       Filename:       tcl3dFile.tcl
#
#       Author:         Paul Obermeier
#
#       Description:    Tcl module with file handling Tcl3D utility procedures.
#
#******************************************************************************

# Procedure to get the location of a writable temporary directory.

proc tcl3dGetTmpDir {} {
    global tcl_platform env

    set tmpDir ""
    # Try different environment variables.
    if { [info exists env(TMP)] && [file isdirectory $env(TMP)] } {
        set tmpDir $env(TMP)
    } elseif { [info exists env(TEMP)] && [file isdirectory $env(TEMP)] } {
        set tmpDir $env(TEMP)
    } elseif { [info exists env(TMPDIR)] && [file isdirectory $env(TMPDIR)] } {
        set tmpDir $env(TMPDIR)
    } else {
        # Last resort. These directories should be available at least.
        switch $tcl_platform(platform) {
            windows {
                if { [file isdirectory "C:/Windows/Temp"] } {
                    set tmpDir "C:/Windows/Temp"
                } elseif { [file isdirectory "C:/Winnt/Temp"] } {
                    set tmpDir "C:/Winnt/Temp"
                }
            }
            unix {
                if { [file isdirectory "/tmp"] } {
                    set tmpDir "/tmp"
                }
            }
        }
    }
    return [file nativename $tmpDir]
}

# Procedure to create a unique temporary directory.
# Returns the name of the created directory.

proc tcl3dCreateTmpDir {} {
    set masterTmpDir [tcl3dGetTmpDir]
    set i 0
    while { 1 } {
        set tmpDirName [format "tcl3d_%05d.tmp" $i]
        set tmpDir [file join $masterTmpDir $tmpDirName]
        # puts "Checking for directory $masterTmpDir"
        if { [file exists $tmpDir] } {
            # Directory already exists. Try to delete it and recreate it first,
            # before attempting to create new number.
            set retVal [catch { file delete -force $tmpDir } ]
            if { $retVal != 0 } {
                incr i
                continue
            }
        }
        file mkdir $tmpDir
        # puts "\tCreating directory $tmpDir"
        return $tmpDir
    }
}

proc tcl3dGenExtName { fileName } {
    if { [info exists      starkit::topdir] && \
         [file isdirectory $starkit::topdir] } {
        set rootDirName [file dirname $::starkit::topdir]
        set extDirName [file join $rootDirName "tcl3dData"]
        return [file join $extDirName [file tail $fileName]]
     } else {
         return $fileName
     }
}

# Procedure to transparently handle files, which may be included in a Starpack.
# You may use this procedure for files (like font files used by FTGL).

proc tcl3dGetExtFile { fileName } {
    # puts "tcl3dGetExtFile $fileName"
    if { [info exists      starkit::topdir] && \
         [file isdirectory $starkit::topdir] } {
        # The script is run from within a Starpack.
        # Now there are 2 possibilities:
        # 1. The requested file is contained within the Starpack.
        #    Then we copy it out to a subdirectory "tcl3dData" of the
        #    directory containing the Starpack.
        # 2. The requested file is not contained within the Starpack.
        #    Then we can return the file name as is.
        if { [string first $starkit::topdir $fileName] == 0 } {
            # puts "Starpack: File inside of Starpack."
            set rootDirName [file dirname $::starkit::topdir]
            set extDirName [file join $rootDirName "tcl3dData"]
            if { ! [file isdirectory $extDirName] } {
                file mkdir $extDirName
            }
            set extFileName [file join $extDirName [file tail $fileName]]
            if { [file exists $extFileName] } {
                # puts "\t$extFileName already exists. No action."
            } else {
                # puts "\tCopy $fileName to $extDirName"
                file copy $fileName $extDirName
            }
            # puts "\tReturning <$extFileName>"
            return $extFileName
        } else {
            # puts "Starpack: File outside of Starpack. Returning <$fileName>"
            return $fileName
        }
    }
    # puts "No Starpack: Returning <$fileName>"
    return $fileName
}

